
def cluster_list_fixture():
    """clusters fixture

    :rtype: dict
    """

    return {
        "cluster_id": "8c4f77ff-849c-456d-a480-c5cb6766c3f2",
        "name": "tamar-ytck1ge",
        "url": "https://52.25.204.103",
        "version": "1.9-dev",
        "attached": False
    }
