{
  "id": "helloworld",
  "cpus": 1.0,
  "mem": 512,
  "instances": 1,
  "cmd": "python3 -m http.server {{port}}",
  "container": {
    "type": "DOCKER",
    "docker": {
      "image": "python:3",
      "network": "HOST"
    }
  }
}
